import logging
from typing import Dict, List, Tuple
from collections.abc import Callable

from .modbus_models import ( MODBUS_VALUE_TYPES, ModbusDeviceInfo, ModbusPointKey, ModbusDevice, ModbusDatapoint, ModbusDatapointKey, ModbusSetpoint, ModbusSetpointKey )

_LOGGER = logging.getLogger(__name__)

class ModbusDeviceAdapter(ModbusDevice):
    _device_info: ModbusDeviceInfo
    _loaded_model: ModbusDevice|None
    
    def __init__(self):
        pass
    
    def instantiate(self, device_info: ModbusDeviceInfo) -> None:
        self._device_info = device_info
        model_to_load = self._translate_to_model(device_info)
        if model_to_load == None:
            raise Exception("Invalid model")
        loaded_model = model_to_load(device_info)
        loaded_model.instantiate(device_info)
        self._loaded_model = loaded_model
     
    def _translate_to_model(self, device_info: ModbusDeviceInfo) -> Callable[[ModbusDeviceInfo], ModbusDevice]|None:
        """Translate the model to the correct device model, must be implemented in the subclass"""
        raise NotImplementedError("Method not implemented")

    def provides_model(self, device_info: ModbusDeviceInfo) -> bool:
        return self._translate_to_model(device_info) is not None
    
    #region ModbusDevice
    
    def _get_loaded_model(self) -> ModbusDevice:
        if self._loaded_model is None:
            raise Exception("No model loaded")
        return self._loaded_model

    def get_datapoint(self, key: ModbusDatapointKey) -> ModbusDatapoint|None:
        return self._get_loaded_model().get_datapoint(key)

    def get_datapoints_for_read(self) -> List[ModbusDatapoint]:
        return self._get_loaded_model().get_datapoints_for_read()

    @property
    def ready(self): return self._loaded_model is not None and self._loaded_model.ready
    @property
    def device_id(self): return self._get_loaded_model().device_id
    @property
    def device_host(self): return self._get_loaded_model().device_host
    @property
    def device_port(self): return self._get_loaded_model().device_port
    @property
    def identification(self): return self._get_loaded_model().identification
    @property
    def manufacturer(self): return self._get_loaded_model().manufacturer
    @property
    def model_name(self): return self._get_loaded_model().model_name
    @property
    def version(self): return self._get_loaded_model().version
    
    
    def get_max_value(self, key: ModbusSetpointKey) -> float|int|None:
        return self._get_loaded_model().get_max_value(key)

    def get_min_value(self, key: ModbusSetpointKey) -> float|int|None:
        return self._get_loaded_model().get_min_value(key)

    def get_setpoint(self, key: ModbusSetpointKey) -> ModbusSetpoint|None:
        return self._get_loaded_model().get_setpoint(key)

    def get_setpoint_step(self, key: ModbusSetpointKey) -> float|int:
        return self._get_loaded_model().get_setpoint_step(key)

    def get_setpoints_for_read(self) -> List[ModbusSetpoint]:
        return self._get_loaded_model().get_setpoints_for_read()
    
    def get_unit_of_measure(self, key: ModbusPointKey) -> str|None:
        return self._get_loaded_model().get_unit_of_measure(key)
    
    def get_value(self, key: ModbusPointKey) -> MODBUS_VALUE_TYPES|None:
        return self._get_loaded_model().get_value(key)
   
    def get_values(self) -> Dict[ModbusPointKey, MODBUS_VALUE_TYPES|None]:
        return self._get_loaded_model().get_values()
    
    def has_value(self, key: ModbusPointKey) -> bool:
        return self._get_loaded_model().has_value(key)

    def provides(self, key: ModbusPointKey) -> bool:
        return self._get_loaded_model().provides(key)

    def set_read(self, key: ModbusPointKey, read: bool) -> bool:
        return self._get_loaded_model().set_read(key, read)

    def set_values(self, kv: List[Tuple[ModbusPointKey, MODBUS_VALUE_TYPES]]) -> Dict[ModbusPointKey, Tuple[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None]]:
        return self._get_loaded_model().set_values(kv)
    
    #endregion ModbusDevice


import concurrent.futures
from dataclasses import dataclass
# import json
import threading
from typing import Any
import pytest

from credentials import Credentials
import logging

from src.modbus_event_connect.micro_nabto.micro_nabto_connection import Request
from src.modbus_event_connect import MODBUS_VALUE_TYPES
from models.micro_nabto_test_models import ModbusTestDatapointKey, ModbusTestMicroNabto
import asyncio

_LOGGER = logging.getLogger(__name__)

@dataclass
class TestData:
    client: ModbusTestMicroNabto
    data = dict[str, Any]()
    credentials: Credentials
    
@pytest.fixture
def testdata():
    #setup
    credentials = Credentials(["username", "hostname"])
    client = ModbusTestMicroNabto()
    # 
    yield TestData(client=client, credentials=credentials)
    #teardown
    client.stop()

async def test_connection_request():
    request = Request(1)
    responsetask = request.wait_for_response()
    request.notify_waiters()
    await responsetask
    
async def test_connection_request_thread():
    request = Request( 1)
    def thread_method():
        asyncio.run(asyncio.sleep(1))
        request.notify_waiters()
    _listen_thread = threading.Thread(target=thread_method)
    _listen_thread.start()
    await request.wait_for_response()
    
async def test_connection_request_asyncio_loop():
    request = Request(1)
    def thread_method():
        asyncio.run(asyncio.sleep(1))
        request.notify_waiters()
    _loop = asyncio.get_event_loop()
    with concurrent.futures.ThreadPoolExecutor() as pool:
        _loop.run_in_executor(pool, thread_method)
    await request.wait_for_response()

async def test_connect(testdata: TestData):
    credentials = testdata.credentials
    client = testdata.client
    success = await client.connect(credentials.username, "DEVICE_ID", credentials.hostname)
    assert success

async def test_datapoint(testdata: TestData):
    credentials = testdata.credentials
    client = testdata.client
    success = await client.connect(credentials.username, "DEVICE_ID", credentials.hostname)
    assert success
    await client.request_datapoint_data()
    event = asyncio.Event()
    def callback(oldval:MODBUS_VALUE_TYPES|None, newval:MODBUS_VALUE_TYPES|None):
        _LOGGER.debug(f"{ModbusTestDatapointKey.MAJOR_VERSION}: {oldval if oldval is not None else 'None'} -> {newval if newval is not None else 'None'}")
        event.set()
    client.subscribe(ModbusTestDatapointKey.MAJOR_VERSION, callback)
    assert await event.wait()



from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Tuple, TypeVar

from .modbus_models import MODBUS_VALUE_TYPES, ModbusDatapoint, ModbusDatapointKey, ModbusPointKey, ModbusSetpoint, ModbusSetpointKey
from .modbus_deviceadapter import ModbusDeviceAdapter

class ModbusEventConnect(ABC):
    _attr_adapter: ModbusDeviceAdapter
    _subscribers: Dict[ModbusPointKey, List[Callable[[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None], None]]] = {}
    """Callable[old_value, new_value]"""
    
    @abstractmethod
    def is_connected(self) -> bool: 
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def stop(self) -> None:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    async def _request_datapoint_data(self, points: List[ModbusDatapoint]) -> List[Tuple[ModbusDatapoint, MODBUS_VALUE_TYPES]]:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    async def _request_setpoint_data(self, points: List[ModbusSetpoint]) -> List[Tuple[ModbusSetpoint, MODBUS_VALUE_TYPES]]:
        raise NotImplementedError("Method not implemented")
        
    @property
    def version(self): return self._attr_adapter.version
    @property
    def device_host(self): return self._attr_adapter.device_host
    @property
    def device_id(self): return self._attr_adapter.device_id
    @property
    def device_port(self): return self._attr_adapter.device_port
    @property
    def identification(self): return self._attr_adapter.identification
    @property
    def manufacturer(self): return self._attr_adapter.manufacturer
    @property
    def model_name(self) -> str: return self._attr_adapter.model_name
    
    async def request_datapoint_data(self) -> None:
        """Request the current value of all subscribed datapoints. All subscribers will be notified of the new value if changed."""
        points = self._attr_adapter.get_datapoints_for_read()
        if len(points) == 0: return
        self._set_values(await self._request_datapoint_data(points))
            
    async def request_setpoint_data(self) -> None:
        """Request the current value of all subscribed setpoints. All subscribers will be notified of the new value if changed."""
        points = self._attr_adapter.get_setpoints_for_read()
        if len(points) == 0: return
        self._set_values(await self._request_setpoint_data(points))
    
    def subscribe(self, key: ModbusDatapointKey|ModbusSetpointKey, update_method: Callable[[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None], None]):
        """
            Subscribe to a datapoint or setpoint value change.
            
            :param key: The key of the datapoint or setpoint to subscribe to.
            :param update_method: The method to call when the value changes. The Callable will receive old_value and new_value as the two inputs.
            """
        if key not in self._subscribers:
            self._subscribers[key] = []
            self._attr_adapter.set_read(key, True)
        self._subscribers[key].append(update_method)
        value = self._attr_adapter.get_value(key)
        if value is not None:
            update_method(None, value)
    
    def unsubscribe(self, key: ModbusDatapointKey|ModbusSetpointKey, update_method: Callable[[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None], None]):
        """Remove a subscription to a datapoint or setpoint value change."""
        subscribers = self._subscribers.get(key)
        if subscribers is None: return
        if update_method in subscribers:
            if len(subscribers) == 1:
                del self._subscribers[key]
                self._attr_adapter.set_read(key, False)
            else: 
                subscribers.remove(update_method)
    
    MODBUS_POINT_TYPE = TypeVar("MODBUS_POINT_TYPE", ModbusDatapoint, ModbusSetpoint)
    def _set_values(self, kv: List[Tuple[MODBUS_POINT_TYPE, MODBUS_VALUE_TYPES]]) -> None:
        result = self._attr_adapter.set_values([(point.key, value) for point, value in kv])
        self._notify_subscribers(result)
    
    def _notify_subscribers(self, kv: Dict[ModbusPointKey, Tuple[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None]]) -> None:
        for key, (old_value, new_value) in kv.items():
            subscribers = self._subscribers.get(key)
            if subscribers is None: return
            for subscriber in subscribers:
                subscriber(old_value, new_value)


from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from enum import StrEnum
from typing import Dict, List, NotRequired, Optional, Tuple, TypedDict

MODBUS_VALUE_TYPES = float|int|str

class ModbusPointKey(StrEnum):
    pass

class ModbusDatapointKey(ModbusPointKey):
    pass

class ModbusSetpointKey(ModbusPointKey):
    pass

class ModbusVersionPointKey(ModbusDatapointKey):
    """Default datapoint keys, which is used to identify the version of the device"""
    MAJOR_VERSION = "major_version"
    MINOR_VERSION = "minor_version"
    PATCH_VERSION = "patch_version"
    
class UOM:
    SECONDS = "seconds"
    MINUTES = "minutes"
    HOURS = "hours"
    DAYS = "days"
    MONTHS = "months"
    YEARS = "years"
    CELSIUS = "celsius"
    BOOL = "bool"
    BITMASK = "bitmask"
    PPM = "ppm"
    """CONCENTRATION PARTS PER MILLION"""
    RPM = "rpm"
    """REVOLUTIONS PER MINUTE"""
    # INT = "int"
    # FLOAT = "float"
    PCT = "percent"
    TEXT = "text"
    UNKNOWN = None
    
class ModbusPointExtras(TypedDict):
    unit_of_measurement: NotRequired[str|None]
    """Unit of measurement for the value, UOM class contains the standard units, defaults to None"""
    always_read: NotRequired[bool|None]
    """Indication if the point should always be read from the device. If not True, the point will only be read if explicitly requested"""
    
@dataclass(kw_only=True)
class ModbusDatapoint:
    key: ModbusDatapointKey
    extra: Optional[ModbusPointExtras|None] = None
    #read
    read_address: int
    signed: bool
    """indication of the data being signed or unsigned (default=True)"""
    divider: Optional[int|None] = None
    """Applied to the register value in the order: 1: divider, 2: offset, 3: modifier"""
    offset: Optional[int|None] = None
    """Applied to the register value in the order: 1: divider, 2: offset, 3: modifier"""
    read_modifier: Optional[Callable[[float|int], float|int]|None] = None
    """Modifier applied to value after it has been parsed by the system. can be used to alter hours to days etc. or round floating values
    Applied to the register value in the order: 1: divider, 2: offset, 3: modifier"""
    read_obj: Optional[int|None] = None
    """default is 0"""

@dataclass(kw_only=True)
class ModbusSetpoint:
    key: ModbusSetpointKey
    extra: Optional[ModbusPointExtras|None] = None
    #read
    read_address: Optional[int|None] = None
    signed: bool
    """indication of the data being signed or unsigned (default=True)"""
    divider: Optional[int|None] = None
    """Applied to the register value in the order: 1: divider, 2: offset, 3: modifier"""
    offset: Optional[int|None] = None
    """Applied to the register value in the order: 1: divider, 2: offset, 3: modifier"""
    read_modifier: Optional[Callable[[float|int], float|int]|None] = None
    """Modifier applied to value after it has been parsed by the system. can be used to alter hours to days etc. or round floating values
    Applied to the register value in the order: 1: divider, 2: offset, 3: modifier"""
    read_obj: Optional[int|None] = None
    """default is 0"""
    #write
    max: int
    """max value in the register"""
    min: int
    """min value in the register"""
    write_address: int|None = None
    step: Optional[int|None] = None
    """step size in register value, if unset will default to the divider"""
    write_modifier: Optional[Callable[[float|int], float|int]|None] = None
    """Modifier applied to value before it has been parsed back to register type. can be used to alter hours to days etc. or round floating values"""
    write_obj: Optional[int|None] = None
    """default is 0"""

    
class ModbusDatapointData:
    point: ModbusDatapoint
    read: bool
    value: MODBUS_VALUE_TYPES|None = None
    unit_of_measurement: str|None = None
    def __init__(self, point: ModbusDatapoint):
        self.point = point
        self.read = point.extra is not None and "always_read" in point.extra and point.extra["always_read"] == True
        self.unit_of_measurement = point.extra["unit_of_measurement"] if point.extra is not None and "unit_of_measurement" in point.extra else None


class ModbusSetpointData:
    point: ModbusSetpoint
    read: bool
    value: MODBUS_VALUE_TYPES|None = None
    unit_of_measurement: str|None = None
    def __init__(self, point: ModbusSetpoint):
        self.point = point
        self.read = point.extra is not None and "always_read" in point.extra and point.extra["always_read"] == True
        self.unit_of_measurement = point.extra["unit_of_measurement"] if point.extra is not None and "unit_of_measurement" in point.extra else None

@dataclass()
class VersionInfo:
    major: int
    minor: int
    patch: int
    
@dataclass(kw_only=True)
class ModbusDeviceIdenfication:
    vendor_name: str|None
    product_code: str|None
    major_minor_revision: str|None
    vendor_url: str|None
    product_name: str|None
    model_name: str|None
    user_application_name: str|None
    
@dataclass(kw_only=True)
class ModbusDeviceInfo:
    device_id: str
    device_host: str
    device_port: int
    version: VersionInfo
    identification: ModbusDeviceIdenfication|None

class ModbusDevice(ABC):
    _ready: bool = False
    _device_info: ModbusDeviceInfo
    
    def __init__(self, device_info: ModbusDeviceInfo) -> None:
        self._device_info = device_info
    
    def instantiate(self, device_info: ModbusDeviceInfo) -> None:
        """
        Sets up the device for usage and raises errors if device has issues.
        Raises:
            ValueError: If the an attribute is not set.
        """
        self._ready = True
    
    @property
    def ready(self) -> bool: return self._ready
    @property
    def device_id(self) -> str: return self._device_info.device_id
    @property
    def device_host(self) -> str: return self._device_info.device_host
    @property
    def device_port(self) -> int: return self._device_info.device_port
    @property
    def identification(self) -> ModbusDeviceIdenfication|None: return self._device_info.identification
    @property
    def version(self) -> VersionInfo: return self._device_info.version
    @property
    @abstractmethod
    def manufacturer(self) -> str:
        raise NotImplementedError("Method not implemented")
    @property
    @abstractmethod
    def model_name(self) -> str:
        raise NotImplementedError("Method not implemented")
    
    @abstractmethod
    def get_datapoint(self, key: ModbusDatapointKey) -> ModbusDatapoint|None:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_datapoints_for_read(self) -> List[ModbusDatapoint]:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_max_value(self, key: ModbusSetpointKey) -> float|int|None:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_min_value(self, key: ModbusSetpointKey) -> float|int|None:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_setpoint(self, key: ModbusSetpointKey) -> ModbusSetpoint|None:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_setpoint_step(self, key: ModbusSetpointKey) -> float|int:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_setpoints_for_read(self) -> List[ModbusSetpoint]:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_unit_of_measure(self, key: ModbusPointKey) -> str|None:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_value(self, key: ModbusPointKey) -> MODBUS_VALUE_TYPES|None:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def get_values(self) -> Dict[ModbusPointKey, MODBUS_VALUE_TYPES|None]:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def has_value(self, key: ModbusPointKey) -> bool:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def provides(self, key: ModbusPointKey) -> bool:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def set_read(self, key: ModbusPointKey, read: bool) -> bool:
        raise NotImplementedError("Method not implemented")
    @abstractmethod
    def set_values(self, kv: List[Tuple[ModbusPointKey, MODBUS_VALUE_TYPES]]) -> Dict[ModbusPointKey, Tuple[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None]]:
        """Sets the values for the keys and returns a list with the old and new values"""
        raise NotImplementedError("Method not implemented")
    
class ModbusDeviceBase(ModbusDevice):
    _attr_manufacturer:str = ""
    """Manufacturer of the device. Must be assigned in the __init__ method"""
    _attr_model_name:str = ""
    """Model name of the device. Must be assigned in the __init__ method"""
    _attr_datapoints: List[ModbusDatapoint]
    """Datapoints for the device. Must be assigned in the __init__ method"""
    _attr_setpoints: List[ModbusSetpoint]
    """Setpoints for the device. Must be assigned in the __init__ method"""
    _attr_default_extras = dict[ModbusPointKey, ModbusPointExtras]()
    """Default extras for the device. Can be assigned in the __init__ method"""
    
    _datapoints = dict[ModbusDatapointKey, ModbusDatapointData]()
    _setpoints = dict[ModbusSetpointKey, ModbusSetpointData]()

    def __init__(self, device_info: ModbusDeviceInfo) -> None:
        super().__init__(device_info)

    @property
    def manufacturer(self) -> str:
        return self._attr_manufacturer
    @property
    def model_name(self) -> str:
        return self._attr_model_name
    
    def instantiate(self, device_info: ModbusDeviceInfo) -> None:
        if not self._attr_manufacturer:
            raise ValueError("Manufacturer not set")
        if not self._attr_model_name:
            raise ValueError("Model name not set")
        if not hasattr(self, '_attr_datapoints'):
            raise ValueError("Datapoints not set")
        if not hasattr(self, '_attr_setpoints'):
            raise ValueError("Setpoints not set")
        
        for point in self._attr_datapoints:
            point.extra = point.extra or self._attr_default_extras.get(point.key)
        self._datapoints = {point.key: ModbusDatapointData(point) for point in self._attr_datapoints}
        
        for point in self._attr_setpoints:
            point.extra = point.extra or self._attr_default_extras.get(point.key)
        self._setpoints = {point.key: ModbusSetpointData(point) for point in self._attr_setpoints}
    
    def get_datapoint(self, key: ModbusDatapointKey) -> ModbusDatapoint | None:
        data = self._datapoints.get(key)
        return data.point if data is not None else None
    
    def get_datapoints_for_read(self) -> List[ModbusDatapoint]:
        return [value.point for value in self._datapoints.values() if value.read]

    def get_max_value(self, key: ModbusSetpointKey) -> float | int | None:
        if self.provides(key):
            point = self._setpoints[key].point
            return ModbusParser.parse_from_modbus_value(point=point, value=point.max)
        return None

    def get_min_value(self, key: ModbusSetpointKey) -> float | int | None:
        if self.provides(key):
            point = self._setpoints[key].point
            return ModbusParser.parse_from_modbus_value(point=point, value=point.min)
        return None

    def get_setpoint(self, key: ModbusSetpointKey) -> ModbusSetpoint | None:
        data = self._setpoints.get(key)
        return data.point if data is not None else None
    
    def get_setpoint_step(self, key: ModbusSetpointKey) -> float|int:
        data = self._setpoints.get(key)
        if data is not None:
            divider = ModbusParser.get_point_divider(data.point)    
            step = ModbusParser.get_point_step(data.point) 
            if divider > 1: return step / divider
            return step
        return 1
    
    def get_setpoints_for_read(self) -> List[ModbusSetpoint]:
        return [value.point for value in self._setpoints.values() if value.read]
    
    def get_unit_of_measure(self, key: ModbusPointKey) -> str | None:
        if isinstance(key, ModbusDatapointKey):
            data = self._datapoints.get(key)
            if data is not None:
                return data.unit_of_measurement
        elif isinstance(key, ModbusSetpointKey):
            data = self._setpoints.get(key)
            if data is not None:
                return data.unit_of_measurement
        return None

    def get_value(self, key: ModbusPointKey) -> MODBUS_VALUE_TYPES|None:
        if isinstance(key, ModbusDatapointKey):
            point = self._datapoints.get(key)
            if point is not None:
                return point.value
        elif isinstance(key, ModbusSetpointKey):
            point = self._setpoints.get(key)
            if point is not None:
                return point.value
        return None

    def get_values(self) -> Dict[ModbusPointKey, MODBUS_VALUE_TYPES|None]:
        datapoints = {key: value.value for key, value in self._datapoints.items() if value.read}
        setpoints = {key: value.value for key, value in self._setpoints.items() if value.read}
        return {**datapoints, **setpoints}
    
    def has_value(self, key: ModbusPointKey) -> bool:
        if isinstance(key, ModbusDatapointKey):
            return key in self._datapoints
        elif isinstance(key, ModbusSetpointKey):
            return key in self._setpoints
        return False

    def provides(self, key: ModbusPointKey) -> bool:
        return key in self._datapoints or key in self._setpoints

    def set_read(self, key: ModbusPointKey, read: bool) -> bool:
        """
        Sets the read state for the point. Returns the new read state.
        If the device sets the always_read to True in the configuration for the key, the read state can never be set to False.
        """
        if key in self._datapoints:
            self._datapoints[key].read = read
            return True
        elif key in self._setpoints:
            data = self._setpoints[key]
            if data.point.read_address is not None:
                data.read = read
                return True
        return False
    
    def _set_value(self, key: ModbusPointKey, value: MODBUS_VALUE_TYPES) -> Tuple[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None]:
        old_value:MODBUS_VALUE_TYPES|None = None
        assigned_value:MODBUS_VALUE_TYPES|None = None
        if key in self._datapoints:
            data = self._datapoints.get(key)
            if data is not None:
                old_value = data.value
                assigned_value = data.value = value
        elif key in self._setpoints:
            data = self._setpoints.get(key)
            if data is not None:
                assigned_value = data.value = value
        return (old_value, assigned_value)
    
    #TODO: Implement this method, to allow us to check version changes in a single go, istead of 3 individual events
    #make sure the vsion change can be overriden by the device, to allow for custom version handling. Ex. add or remove features based on version.
    def set_values(self, kv: List[Tuple[ModbusPointKey, MODBUS_VALUE_TYPES]]) -> Dict[ModbusPointKey, Tuple[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None]]:
        result = dict[ModbusPointKey, Tuple[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None]]()
        for key, value in kv:
            old_value, new_value = self._set_value(key, value)
            result[key] = (old_value, new_value)
        
        self._update_if_version_changed(result)
        return result
    
    def _update_if_version_changed(self, valuesset:Dict[ModbusPointKey, Tuple[MODBUS_VALUE_TYPES|None, MODBUS_VALUE_TYPES|None]]) -> None:
        major = valuesset.get(ModbusVersionPointKey.MAJOR_VERSION)
        minor = valuesset.get(ModbusVersionPointKey.MINOR_VERSION)
        patch = valuesset.get(ModbusVersionPointKey.PATCH_VERSION)
        
        if major is not None or minor is not None or patch is not None:
            major_val = major[1] if major is not None and major[1] is not None else 0
            minor_val = minor[1] if minor is not None and minor[1] is not None else 0
            patch_val = patch[1] if patch is not None and patch[1] is not None else 0
            self._device_info.version = VersionInfo(major=int(major_val), minor=int(minor_val), patch=int(patch_val))
            
        
class MODIFIER:
    @staticmethod
    def flip_bool(value:float|int) -> float|int:
        """Flips the true/false state 
        - 1 -> 0
        - 0 -> 1"""
        return 1-value
    
    @staticmethod
    def seconds_to_minutes(value:float|int) -> float|int:
        return round(value/60)
    
    @staticmethod
    def hours_to_days(value:float|int) -> float|int:
        return round(value/24)
    
class ModbusParser:
    @staticmethod
    def parse_to_modbus_value(point:ModbusSetpoint, value: float|int) -> int:
        divider = ModbusParser.get_point_divider(point)
        offset = ModbusParser.get_point_offset(point)
        modifier = ModbusParser.get_point_write_modifier(point)
        new_value:float|int = value
        if modifier is not None: new_value = modifier(new_value)
        if divider > 1: new_value *= divider
        if offset != 0: new_value -= offset 
        return int(new_value) #cast to int, modbus writes only accept an int

    @staticmethod
    def parse_from_modbus_value(point:ModbusDatapoint|ModbusSetpoint, value: int) -> float|int:
        divider = ModbusParser.get_point_divider(point)
        offset = ModbusParser.get_point_offset(point)
        modifier = ModbusParser.get_point_read_modifier(point)
        new_value:float|int = value
        if offset != 0: new_value += offset 
        if divider > 1: new_value /= divider
        if modifier is not None: new_value = modifier(new_value)
        return new_value

    @staticmethod
    def get_point_divider(point:ModbusDatapoint|ModbusSetpoint) -> int: 
        return 1 if point.divider is None else point.divider
    @staticmethod
    def get_point_offset(point:ModbusDatapoint|ModbusSetpoint) -> int: 
        return 0 if point.offset is None else point.offset
    @staticmethod
    def get_point_read_address(point:ModbusDatapoint|ModbusSetpoint) -> int|None: 
        return point.read_address
    @staticmethod
    def get_point_read_obj(point:ModbusDatapoint|ModbusSetpoint) -> int: 
        return 0 if point.read_obj is None else point.read_obj
    @staticmethod
    def get_point_write_address(point:ModbusSetpoint) -> int|None: 
        return point.write_address
    @staticmethod
    def get_point_write_obj(point:ModbusSetpoint) -> int: 
        return 0 if point.write_obj is None else point.write_obj
    @staticmethod
    def get_point_signed(point:ModbusDatapoint|ModbusSetpoint) -> bool: 
        return point.signed
    @staticmethod
    def get_point_step(point:ModbusSetpoint) -> int: 
        return 1 if point.step is None else point.step
    @staticmethod
    def get_point_max(point:ModbusSetpoint) -> int: 
        return point.max
    @staticmethod
    def get_point_min(point:ModbusSetpoint) -> int: 
        return point.min
    @staticmethod
    def get_point_read_modifier(point:ModbusDatapoint|ModbusSetpoint) -> Callable[[float|int], float|int]|None: 
        return None if point.read_modifier is None else point.read_modifier
    @staticmethod
    def get_point_write_modifier(point: ModbusSetpoint) -> Callable[[float|int], float|int]|None: 
        return None if point.write_modifier is None else point.write_modifier


from .modbus_event_connect import ( ModbusEventConnect )
from .modbus_deviceadapter import ( ModbusDeviceAdapter )
from .modbus_models import ( 
        ModbusDevice, 
        ModbusDeviceBase, 
        ModbusDeviceInfo,
        ModbusDatapoint, 
        ModbusDatapointKey, 
        ModbusParser,
        ModbusPointKey, 
        ModbusSetpoint, 
        ModbusSetpointKey, 
        ModbusVersionPointKey,
        MODBUS_VALUE_TYPES,
        MODIFIER,
        UOM,
        VersionInfo,
        )
from .micro_nabto.micro_nabto_connection import ( MicroNabtoModbusDeviceInfo )
from .micro_nabto.micro_nabto_event_connect import ( MicroNabtoEventConnect )
from .modbus_tcp.modbus_tcp_event_connect import ( ModbusTCPEventConnect )

__version__ = "0.1.2"
__all__ = [
    "ModbusDevice",
    "ModbusDeviceAdapter",
    "ModbusDeviceBase",
    "ModbusDeviceInfo",
    "ModbusDatapoint",
    "ModbusDatapointKey",
    "ModbusParser",
    "ModbusPointKey",
    "ModbusEventConnect",
    "ModbusSetpoint",
    "ModbusSetpointKey",
    "ModbusVersionPointKey",
    "MODBUS_VALUE_TYPES",
    "MODIFIER",
    "UOM",
    "MicroNabtoEventConnect",
    "MicroNabtoModbusDeviceInfo",
    "ModbusTCPEventConnect",
    "VersionInfo",
]